@@ -472,9 +472,17 @@ class TestWaitConditions(unittest.TestCase):
             self._assert_inclusive_range(fn(make_retry_state(8, 0)), 0, 60.0)
             self._assert_inclusive_range(fn(make_retry_state(9, 0)), 0, 60.0)
 
-        fn = tenacity.wait_random_exponential(10, 5)
+        # max wait
+        max_wait = 5
+        fn = tenacity.wait_random_exponential(10, max_wait)
         for _ in range(1000):
-            self._assert_inclusive_range(fn(make_retry_state(1, 0)), 0.00, 5.00)
+            self._assert_inclusive_range(fn(make_retry_state(1, 0)), 0.00, max_wait)
+
+        # min wait
+        min_wait = 5
+        fn = tenacity.wait_random_exponential(min=min_wait)
+        for _ in range(1000):
+            self._assert_inclusive_range(fn(make_retry_state(1, 0)), min_wait, 5)
 
         # Default arguments exist
         fn = tenacity.wait_random_exponential()
